import cntk as C
from cntkx.ops import cumsum, hardmax, erf, batchmatmul, scalar, gelu, gelu_fast, floor_division, remainder
import numpy as np
from numpy.testing import assert_equal
import pytest


def test_floor_division():
    x = [-3, 1, 2, 3, 4, 5.2]
    y = [2, 2, 2, 2, 2, 2]
    a = C.constant(x)
    b = C.constant(y)

    desired = [i // j for i, j in zip(x, y)]  # [-2, 0, 1, 1, 2, 2]
    result = floor_division(a, b).eval().tolist()
    assert result == desired


def test_remainder():
    x = [-3, 1, 2, 3, 4, 3, 5.123]
    y = [2, 2, 2, 2, 2, -2, -1.234]
    a = C.constant(x)
    b = C.constant(y)

    desired = [i % j for i, j in zip(x, y)]  # [1, 1, 0, 1, 0, -1, ...]
    result = remainder(a, b).eval().tolist()
    assert pytest.approx(result) == desired


def test_scalar():
    a = C.input_variable((3, 64, 64))
    b = scalar(a)

    assert b.shape == (1,)

    n = np.random.random((10, 3, 64, 64)).astype(np.float32)
    b.eval({a: n})


def test_cumsum():
    a = C.input_variable(5)
    b = cumsum(a)

    n = np.array([1, 2, 3, 4, 5]).astype(np.float32)[None, ...]
    results = b.eval({a: n})
    assert_equal(results[0], n.cumsum())


def test_hardmax():
    a = C.input_variable((3, 5))

    n = np.array([[0.2, 0.2, 0.3, 0.2, 0.2],
                  [0.2, 0.4, 0.3, 0.2, 0.2],
                  [0.5, 0.2, 0.3, 0.2, 0.2]]).astype(np.float32)[None, ...]

    m = np.array([[0, 0, 1, 0, 0],
                  [0, 1, 0, 0, 0],
                  [1, 0, 0, 0, 0]])

    results = hardmax(a).eval({a: n})

    assert_equal(results[0], m)

    n = np.array([[0.2, 0.2, 0.3, 0.2, 0.2],
                  [0.2, 0.4, 0.3, 0.2, 0.2],
                  [0.5, 0.2, 0.3, 0.2, 0.2]]).astype(np.float32)[None, ...]

    m = np.array([[0, 0, 0, 0, 0],
                  [0, 0, 0, 0, 0],
                  [1, 0, 0, 0, 0]])

    results = hardmax(a, axis=None).eval({a: n})

    assert_equal(results[0], m)


def test_erf():
    a = C.input_variable(1)
    b = erf(a)

    n = np.array([[0], [0.02], [0.04], [0.06], [0.08], [0.1]], dtype=np.float32)
    ans = np.array([[0], [0.022564575], [0.045111106], [0.067621594], [0.090078126], [0.112462916]])
    results = b.eval({a: n})

    np.testing.assert_almost_equal(np.array(results), ans, decimal=6)

    a = C.input_variable(6)
    b = erf(a)

    n = np.array([[0, 0.02, 0.04, 0.06, 0.08, 0.1], ], dtype=np.float32)
    ans = np.array([[0, 0.022564575, 0.045111106, 0.067621594, 0.090078126, 0.112462916], ])
    results = b.eval({a: n})

    np.testing.assert_almost_equal(np.array(results), ans, decimal=6)


def test_gelu():
    a = C.input_variable(10)
    b = gelu(a)

    assert b.shape == (10,)

    n = np.random.random((10, 10)).astype(np.float32)
    b.eval({a: n})


def test_gelu_fast():
    a = C.input_variable(10)
    b = gelu_fast(a)

    assert b.shape == (10,)

    n = np.random.random((10, 10)).astype(np.float32)
    b.eval({a: n})


def test_seq_batchmatmul0():
    """ sequence axis present, left operand is matrix and right operand is vector """
    dynamic_batch = 2
    seq_len = 4
    matmul_batch = 3

    batch_shape = (dynamic_batch, seq_len, matmul_batch)

    left_operand_shape = (5, 2)
    right_operand_shape = (2, 1)
    final_shape = (5, 1)

    n = np.random.random(batch_shape + left_operand_shape).astype(np.float32)
    m = np.random.random(batch_shape + right_operand_shape).astype(np.float32)

    a = C.sequence.input_variable((matmul_batch,) + left_operand_shape)
    b = C.sequence.input_variable((matmul_batch,) + right_operand_shape)

    c = batchmatmul(a, b)
    assert c.shape == (matmul_batch, ) + final_shape

    desired = n @ m
    result = c.eval({a: n, b: m})
    result = np.array(result)

    np.testing.assert_almost_equal(result, desired, decimal=7)


def test_seq_batchmatmul1():
    """ sequence axis present, left operand is matrix and right operand is matrix """
    dynamic_batch = 2
    seq_len = 4
    matmul_batch = 3

    batch_shape = (dynamic_batch, seq_len, matmul_batch)

    left_operand_shape = (5, 2)
    right_operand_shape = (2, 3)
    final_shape = (5, 3)

    n = np.random.random(batch_shape + left_operand_shape).astype(np.float32)
    m = np.random.random(batch_shape + right_operand_shape).astype(np.float32)

    a = C.sequence.input_variable((matmul_batch,) + left_operand_shape)
    b = C.sequence.input_variable((matmul_batch,) + right_operand_shape)

    c = batchmatmul(a, b)
    assert c.shape == (matmul_batch, ) + final_shape

    desired = n @ m
    result = c.eval({a: n, b: m})
    result = np.array(result)

    np.testing.assert_almost_equal(result, desired, decimal=7)


def test_seq_batchmatmul2():
    """ sequence axis present, left operand is tensor and right operand is matrix """
    dynamic_batch = 2
    seq_len = 4
    matmul_batch = 3

    batch_shape = (dynamic_batch, seq_len, matmul_batch)

    left_operand_shape = (5, 6, 2)
    right_operand_shape = (2, 3)
    final_shape = (5, 6, 3)

    n = np.random.random(batch_shape + left_operand_shape).astype(np.float32)
    m = np.random.random(batch_shape + right_operand_shape).astype(np.float32)

    a = C.sequence.input_variable((matmul_batch, ) + left_operand_shape)
    b = C.sequence.input_variable((matmul_batch, ) + right_operand_shape)

    c = batchmatmul(a, b)
    assert c.shape == (matmul_batch, ) + final_shape

    desired_packed = n.reshape(batch_shape + (-1, 2)) @ m
    desired = desired_packed.reshape(batch_shape + final_shape)
    result = c.eval({a: n, b: m})
    result = np.array(result)

    np.testing.assert_almost_equal(result, desired, decimal=7)


def test_seq_batchmatmul3():
    """ sequence axis present, left operand is matrix and right operand is tensor """
    dynamic_batch = 2
    seq_len = 4
    matmul_batch = 3

    batch_shape = (dynamic_batch, seq_len, matmul_batch)

    left_operand_shape = (5, 2)
    right_operand_shape = (2, 6, 3)
    final_shape = (5, 6, 3)

    n = np.random.random(batch_shape + left_operand_shape).astype(np.float32)
    m = np.random.random(batch_shape + right_operand_shape).astype(np.float32)

    a = C.sequence.input_variable((matmul_batch, ) + left_operand_shape)
    b = C.sequence.input_variable((matmul_batch, ) + right_operand_shape)

    c = batchmatmul(a, b, output_rank=2)
    assert c.shape == (matmul_batch, ) + final_shape

    desired_packed = n @ m.reshape(batch_shape + (2, -1))
    desired = desired_packed.reshape(batch_shape + final_shape)
    result = c.eval({a: n, b: m})
    result = np.array(result)

    np.testing.assert_almost_equal(result, desired, decimal=7)


def test_seq_batchmatmul4():
    """ No sequence axis, left operand is matrix, right operand is matrix """
    dynamic_batch = 2
    matmul_batch = 3

    batch_shape = (dynamic_batch, matmul_batch)

    inner_dim = 2
    left_operand_shape = (5, inner_dim)
    right_operand_shape = (inner_dim, 6)
    final_shape = (5, 6)

    n = np.random.random(batch_shape + left_operand_shape).astype(np.float32)
    m = np.random.random(batch_shape + right_operand_shape).astype(np.float32)

    a = C.input_variable((matmul_batch,) + left_operand_shape)
    b = C.input_variable((matmul_batch,) + right_operand_shape)

    c = batchmatmul(a, b, output_rank=1)
    assert c.shape == (matmul_batch,) + final_shape

    desired_packed = n @ m
    desired = desired_packed
    result = c.eval({a: n, b: m})
    result = np.array(result)

    np.testing.assert_almost_equal(result, desired, decimal=7)


def test_seq_batchmatmul5():
    """ No sequence axis, left operand is tensor, right operand is matrix """
    dynamic_batch = 2
    matmul_batch = 3

    batch_shape = (dynamic_batch, matmul_batch)

    inner_dim = 2
    left_operand_shape = (5, 7, inner_dim)
    right_operand_shape = (inner_dim, 6)
    final_shape = (5, 7, 6)

    n = np.random.random(batch_shape + left_operand_shape).astype(np.float32)
    m = np.random.random(batch_shape + right_operand_shape).astype(np.float32)

    a = C.input_variable((matmul_batch,) + left_operand_shape)
    b = C.input_variable((matmul_batch,) + right_operand_shape)

    c = batchmatmul(a, b, output_rank=1)
    assert c.shape == (matmul_batch,) + final_shape

    desired_packed = n.reshape(batch_shape + (-1, 2)) @ m
    desired = desired_packed.reshape(batch_shape + final_shape)
    result = c.eval({a: n, b: m})
    result = np.array(result)

    np.testing.assert_almost_equal(result, desired, decimal=7)


def test_seq_batchmatmul6():
    """ No sequence axis, left operand is matrix, right operand is tensor """
    dynamic_batch = 2
    matmul_batch = 3

    batch_shape = (dynamic_batch, matmul_batch)

    inner_dim = 2
    left_operand_shape = (5, inner_dim)
    right_operand_shape = (inner_dim, 6, 3)
    final_shape = (5, 6, 3)

    n = np.random.random(batch_shape + left_operand_shape).astype(np.float32)
    m = np.random.random(batch_shape + right_operand_shape).astype(np.float32)

    a = C.input_variable((matmul_batch,) + left_operand_shape)
    b = C.input_variable((matmul_batch,) + right_operand_shape)

    c = batchmatmul(a, b, output_rank=2)
    assert c.shape == (matmul_batch, ) + final_shape

    desired_packed = n @ m.reshape(batch_shape + (inner_dim, -1))
    desired = desired_packed.reshape(batch_shape + final_shape)
    result = c.eval({a: n, b: m})
    result = np.array(result)

    np.testing.assert_almost_equal(result, desired, decimal=7)


def test_seq_batchmatmul7():
    """
    sequence axis present with samples of uneven sequence length
    left operand is matrix, right operand is matrix
    """

    # uneven sequence length between batches
    n = [np.random.random((4,   3,   5, 2)).astype(np.float32),
         np.random.random((2,   3,   5, 2)).astype(np.float32),
         np.random.random((7,   3,   5, 2)).astype(np.float32)]

    m = [np.random.random((4,   3,   2, 7)).astype(np.float32),
         np.random.random((2,   3,   2, 7)).astype(np.float32),
         np.random.random((7,   3,   2, 7)).astype(np.float32)]

    a = C.sequence.input_variable((3, 5, 2))
    b = C.sequence.input_variable((3, 2, 7))

    c = batchmatmul(a, b)
    assert c.shape == (3, 5, 7)

    desired_packed = [nn @ mm for nn, mm in zip(n, m)]
    desired_results = desired_packed
    results = c.eval({a: n, b: m})

    assert len(results) == len(desired_results)

    for result, desired in zip(results, desired_results):
        np.testing.assert_almost_equal(result, desired, decimal=7)



import cntk as C
from cntkx.ops import gaussian_mdn_coeff
from math import pi


@C.typemap
def gaussian_mdn_loss(output_vector, target_vector, nmix: int, ndim: int):
    """
    Loss function for gaussian mixture density network. Usually used for regression problems.
    Mixture density networks are useful when trying to represent arbitrary conditional probabilities
    the same way a conventional neural network can represent arbitrary functions.

    Example:
        ndim, nmix = 1, 3
        input_tensor = C.input_variable(1, name="input_tensor")
        target_tensor = C.input_variable(1, name="target_tensor")

        # model
        inner = Dense(50, activation=C.relu)(input_tensor)
        inner = Dense(50, activation=C.relu)(inner)
        prediction_tensor = Dense((ndim + 2) * nmix, activation=None)(inner)

        loss = gaussian_mdn_loss(prediction_tensor, target_tensor, nmix=nmix, ndim=ndim)

    Arguments:
        output_vector: network output
        target_vector: ground truths (typically a continuous variable)
        nmix (int): number of mixtures
        ndim (int): number of dimensions in a gaussian kernel

    Returns:
        :class:`~cntk.ops.functions.Function`
    """

    @C.typemap
    def gaussian_mdn_phi(target, mu, sigma, ndim: int):
        """
        Calculates phi between the target tensor and the network prediction
        Does not assumes independence between components of target.

        Arguments:
            target: target tensor with shape (ndim, )
            mu: means of gaussian mdn with shape (nmix, ndim)
            sigma: sigma of gaussian mdn
            nmix (int): number of mixtures
            ndim (int): number of dimensions in gaussian

        Returns:
            :class:`~cntk.ops.functions.Function`
        """
        if not len(mu.shape) == 2:
            raise ValueError("mu {0} must have shape (nmix, ndim)".format(mu.shape))

        t = C.expand_dims(target, axis=0)

        exp_term = C.exp(C.negate(C.square(C.reduce_l2(t - mu, axis=-1)) / (2 * C.square(sigma))))
        factor = C.reciprocal((2 * pi) ** (ndim / 2) * C.pow(sigma, ndim))
        return factor * exp_term

    alpha, mu, sigma = gaussian_mdn_coeff(output_vector, nmix=nmix, ndim=ndim)
    phi = gaussian_mdn_phi(target_vector, mu, sigma, ndim=ndim)
    loss = C.negate(C.log(C.clip(C.reduce_sum(alpha * phi, axis=0), 1e-10, 1e10)))
    return loss


@C.typemap
def focal_loss_with_softmax(output_vector, target_vector, alpha=1, gamma=2., axis=-1, name=''):
    """
    CNTK multi-class implementation of focal loss from "Focal Loss for Dense Object Detection" by Tsung-Yi Lin et al.

    Focal loss add a factor (1 - p) ^ gamma to the standard cross entropy criterion. Setting gamma > 0 reduces the
    relative loss for well-classified examples (p > .5), putting more focus on hard, misclassified examples.
    Focal loss enables the training of highly accurate dense object detectors in the presence of vast
    numbers of easy background examples or dataset with extreme class imbalance (e.g. 1:1000).

    This implementation will work in semantic segmentation of images i.e. output can
    be a rank 2 tensor of shape (num_classes, row, col)

    Maths:
        Focal Loss = - alpha * (1 - p) ^ gamma * log ( p )

    Example:
        Cx.focal_loss_with_softmax([[0, 0, 0.8, 0.2]], [[0, 0, 1, 0]]).eval()
        array([[0.31306446]], dtype=float32)

    Arguments:
        output_vector: the unscaled computed output values from the network. Can be
          from shape (num_classes,) for classification up to shape (num_classes, row, col) for semantic segmentation
          of images.
        target_vector: usually it is one-hot vector where the hot bit
         corresponds to the label index. But it can be any probability
         distribution over the labels.
        alpha (float): sacling factor. weight assigned to rare classes.
          should slightly decrease as gamma increase. (defaults 1)
        gamma (float): Larger gamma reduces relative loss for well-classified examples.
          Recommended range [0.5, 5] (Default 2.)
        axis (int or :class:`~cntk.axis.Axis`, optional): if given, focal loss will be computed
                along this axis
        name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`

    """
    prob = C.softmax(output_vector, axis=axis)
    log_prob = target_vector * C.log(prob)  # cross entropy with softmax

    factor = C.pow(1 - prob, gamma)

    return C.negate(alpha * C.reduce_sum(factor * log_prob, axis=axis), name=name)


@C.typemap
def binary_focal_loss(output, target, alpha=1., gamma=2., name=''):
    """
    CNTK binary class implementation of focal loss from "Focal Loss for Dense Object Detection" by Tsung-Yi Lin et al.

    Focal loss add a factor (1 - p) ^ gamma to the standard cross entropy criterion. Setting gamma > 0 reduces the
    relative loss for well-classified examples (p > .5), putting more focus on hard, misclassified examples.
    Focal loss enables the training of highly ccurate dense object detectors in the presence of vast
    numbers of easy background examples or dataset with extreme class imbalance (e.g. 1:1000).

    This implementation will work in semantic segmentation of images i.e. output can
    be a rank 2 tensor of shape (row, col). Output will be correct even in edge case where entire image is background.

    Maths:
        Focal Loss = - alpha * (1 - p) ^ gamma * log ( p )

    Arguments:
        output: the computed posterior probability from the network (typ. a ``sigmoid``). Can be
          from shape (1,) for simple classification up to shape (row, col) for semantic segmentation of images.
        target: ground-truth label, 0 or 1
        alpha (float): sacling factor. weight assigned to rare classes.
          should slightly decrease as gamma increase. (defaults 1)
        gamma (float): Larger gamma reduces relative loss for well-classified examples.
          Recommended range [0.5, 5] (Default 2.)
        axis (int or :class:`~cntk.axis.Axis`, optional): if given, focal loss will be computed
                along this axis
        name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`

    """
    logprobA = target * C.log(output)
    logprobB = (1 - target) * C.log(1 - output)

    factorA = C.pow(1 - output, gamma)
    factorB = C.pow(output, gamma)

    return C.negate(alpha * (factorA * logprobA + factorB * logprobB), name=name)


def cross_entropy_with_softmax(output_vector, target_vector, axis=-1, label_smoothing=0., name=''):
    """ Adds label smoothing regularisation to C.cross_entropy_with_softmax

    Label smoothing regularisation prevents the largest logit from becoming much larger than all others.

    Classical cross entropy maximises log-likelihood of the correct label. This can cause overfitting.
    If the model learns to assign full probability to the ground-truth label for each training example, it
    is not guaranteed to generalize. Second, it encourages the differences between the largest logit and all others to
    become large, and this, combined with the bounded gradient, reduces the ability of the model to adapt.
    Intuitively, this happens because the model becomes too confident about its predictions.

    To solve this, we introduce label smoothing.

    For more details on label smoothing, please refer to Rethinking the Inception Architecture for Computer Vision
    by Szegedy et al. (https://arxiv.org/abs/1512.00567)

    Arguments:
        output_vector: the unscaled computed output values from the network
        target_vector: usually it is one-hot vector where the hot bit
         corresponds to the label index. But it can be any probability
         distribution over the labels.
        axis (int or :class:`~cntk.axis.Axis`, optional): if given, cross entropy will be computed
                along this axis
        label_smoothing (float): a small label smoothing constant. As a guideline, in ImageNet training of 1000 classes,
            label_smoothing can be set to 0.1
        name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`

    """
    if label_smoothing == 0.:
        return C.cross_entropy_with_softmax(output_vector, target_vector, axis=axis, name=name)
    else:
        k = output_vector.shape[axis]
        target_vector = (1 - label_smoothing) * target_vector + label_smoothing / k
        return C.cross_entropy_with_softmax(output_vector, target_vector, axis=axis, name=name)


def generalised_robust_barron_loss(output_vector, target_vector, alpha: float, scale: float = None):
    """ Generalised robust loss that can be used as replacement to either l1 or L2 norm or any regression task.

    By introducing robustness as a continuous parameter, our loss function allows algorithms
    built around robust loss minimization to be generalized, which improves performance on
    basic vision tasks such as registration and clustering.

    This implements the rho(x, \alpha, c) function described in
    "A General and Adaptive Robust Loss Function", Jonathan T. Barron, https://arxiv.org/abs/1701.03077
    View is video on this talk https://www.youtube.com/watch?v=BmNKbnF69eY

    The approximate form of the loss which is faster to compute, but inaccurate
    when the difference between `output_vector` and `target_vector`  and alpha are near zero.

    Different alpha allows you to get various forms of robust loss:
        alpha = 2: L2 loss
        alpha = 1: pseudo-Huber loss
        alpha = 0: Cauchy loss
        alpha = -2: Geman-McClure loss
        alpha = -inf: Welsch loss

    The effect of outlier diminishes as alpha becomes more negative, and as alpha approaches -inf an
    outlier whose residual magnitude is larger than 3 * scale is almost completely ignored.

    This generalised loss for all alpha has the following properties:
        Monotonic (useful for annealing)
        Smooth wrt inputs and alpha
        Bounded first and second derivatives

    Note:
        this is NOT the adaptive implementation of barron loss, which is slightly more complicated to implement,
        since it uses cubic spline interpolation to estimate the partition function.

        I may implement it once i have time.

    Arguments:
        output_vector: tensor (assumes last axis is the dimension axis where the loss is to be computed from)
        target_vector: tensor
        alpha (float): controls the form that the loss takes.The effect of outlier diminishes as alpha becomes more
          negative, and as alpha approaches -inf an outlier whose residual magnitude is larger than 3 * scale
          is almost completely ignored.
        scale (float): controls the size of the loss's quadratic bowl near x = 0, the derivative of the loss is
          approximately linear when |x| < scale. If scale is set as None, it will be learned.

    Returns:
        :class:`~cntk.ops.functions.Function`
    """
    if scale is None:
        scale = C.Parameter(shape=(output_vector.shape[-1],), init=C.glorot_uniform())

    # alpha = C.Parameter(shape=(output_vector.shape[-1],), init=C.glorot_uniform())
    # alpha_negative = C.less(alpha, 0)
    #
    # x = output_vector - target_vector
    #
    # b = C.abs(alpha - 2) + epsilon
    # d = C.element_select(alpha_negative, alpha - epsilon, alpha + epsilon)
    # loss = (b / d) * (C.pow(C.square(x / scale) / b + 1., 0.5 * d) - 1.)

    x = output_vector - target_vector

    # Avoids the 3 singularity at alpha == 2, alpha == 0, and alpha == -inf
    if alpha == 2:
        loss = C.square(x / scale)  # L2 loss
    elif alpha == 0:
        loss = C.log(0.5 * C.square(x / scale) + 1)  # Cauchy loss
    elif alpha <= -1e8:
        loss = 1 - C.exp(-0.5 * C.square(x / scale))  # welsch loss
    else:
        alpha_offset = abs(alpha - 2)
        factor = alpha_offset / alpha

        core = C.pow(C.square(x / scale) / alpha_offset + 1, alpha / 2)
        loss = factor * (core - 1)

    return loss



import cntk as C
from cntkx.layers import QRNN, Recurrence, IndyLSTM, IndRNN, WeightDroppedLSTM
from cntk.layers import Dense, LSTM
import numpy as np
import random
import time
from pprint import pprint
from operator import mul
from functools import reduce


def generate_variable_10(nb_samples=1000, dim=3):
    """
    Generate a dataset of sequences where sequence length is between 5 to 20. Sequence will have values 1 or 0.
    Target will be the sum of the sequence.
    """
    r = [random.randint(5, 20) for __ in range(nb_samples)]
    x = [np.random.randint(2, size=(t, dim)).astype(np.float32) for t in r]
    y = np.array([arr.sum() for arr in x])[..., None]
    return x, y


def model_qrnn(input_tensor, hidden_dim):
    hidden = QRNN(window=2, hidden_dim=hidden_dim)(input_tensor)
    prediction = Dense(1)(C.sequence.last(hidden))
    return prediction


def model_lstm(input_tensor, hidden_dim):
    hidden = Recurrence(LSTM(hidden_dim))(input_tensor)
    prediction = Dense(1)(C.sequence.last(hidden))
    return prediction


def model_wdlstm(input_tensor, hidden_dim, dropout):
    hidden = Recurrence(WeightDroppedLSTM(hidden_dim, dropout))(input_tensor)
    prediction = Dense(1)(C.sequence.last(hidden))
    return prediction

def model_indy_lstm(input_tensor, hidden_dim):
    hidden = Recurrence(IndyLSTM(hidden_dim))(input_tensor)
    prediction = Dense(1)(C.sequence.last(hidden))
    return prediction


def model_ind_rnn(input_tensor, hidden_dim):
    hidden = Recurrence(IndRNN(hidden_dim, C.relu))(input_tensor)
    prediction = Dense(1)(C.sequence.last(hidden))
    return prediction


n_epoch = 50
minibatch_size = 30

input_dim = 2
x, y = generate_variable_10(nb_samples=1000, dim=input_dim)
test_x, test_y = generate_variable_10(nb_samples=100, dim=input_dim)

input_tensor = C.sequence.input_variable(input_dim)
target_tensor = C.input_variable(1)

p_lstm = model_lstm(input_tensor, hidden_dim=100)
p_wdlstm = model_wdlstm(input_tensor, hidden_dim=100, dropout=0.1)
p_qrnn = model_qrnn(input_tensor, hidden_dim=800)
p_indy_lstm = model_indy_lstm(input_tensor, hidden_dim=200)
p_ind_rnn = model_ind_rnn(input_tensor, hidden_dim=600)

predictions = [p_lstm, p_wdlstm, p_qrnn, p_indy_lstm, p_ind_rnn]
block_names = ['lstm', 'wdlstm', 'qrnn', 'IndyLSTM', 'IndRNN']

performance = []
for block_name, prediction in zip(block_names, predictions):
    loss = C.squared_error(prediction, target_tensor)
    adam = C.adam(prediction.parameters, 0.1, 0.912)
    trainer = C.Trainer(prediction, (loss,), [adam])

    num_parameters = sum([reduce(mul, p.shape + (1,)) for p in prediction.parameters])

    history = []
    start = time.time()

    for epoch in range(n_epoch):

        for i in range(0, len(x), minibatch_size):
            lbound, ubound = i, i + minibatch_size
            x_mini = x[lbound:ubound]
            y_mini = y[lbound:ubound]
            trainer.train_minibatch({input_tensor: x_mini,
                                     target_tensor: y_mini})

            history.append(trainer.previous_minibatch_loss_average)
            # print(f"loss: {history[-1]}")

    duration = time.time() - start
    print(f"{block_name} training completed in {duration}s")

    loss_result = sum(loss.eval({input_tensor: test_x, target_tensor: test_y}))

    block_performance = (block_name, duration, loss_result, num_parameters)
    performance.append(block_performance)

for block_name, duration, loss_result, num_parameters in performance:
    print(f"name: {block_name}, duration: {duration}, loss: {loss_result}, parameter_count: {num_parameters}")


import cntk as C
import cntkx as Cx
import numpy as np
from cntk.layers import ResNetBlock
from cntkx.layers import PreTrainedBertEmbeddings, PositionwiseFeedForward, Dense, PretrainedBertPooler, Recurrence
from cntkx.layers import LayerNormalization
from cntk.default_options import default_override_or
from cntk.layers.blocks import _inject_name


def LinearAttention(hidden_dim: int, model_dim: int,
                    key_init=default_override_or(C.glorot_uniform()), key_init_bias=default_override_or(0),
                    query_init=default_override_or(C.glorot_uniform()), query_init_bias=default_override_or(0),
                    value_init=default_override_or(C.glorot_uniform()), value_init_bias=default_override_or(0),
                    name=''):
    """ Attention model that is linear in time and memory complexity.
    This is a huge improvement from standard softmax attention models or self-attention
    where the time and memory complexity is quadratic in sequence length.

    This is especially significant since cntk doesn't have any build-in checkpointing functionality
    that saves gpu memory and hence allow the training of Transformer models. With this attention,
    it becomes possible to do transformer training on cntk.

    This implementation addresses the limitation of attentions by express the attention
    as a linear dot-product of kernel feature maps and made use of the associativity property of matrix products.

    When query, key and value are all the same, it becomes self-attention.

    For more details refer to "Transformers are RNNs:Fast Autoregressive Transformers with Linear Attention" by
    Katharopoulos et al. (https://arxiv.org/abs/2006.16236)

    Note:
        Key and value must have the same sequence length

    Example:
        a = C.sequence.input_variable(24)
        b = LinearAttention(hidden_dim=32, model_dim=24)(a, a, a)

        assert b.shape == (32, )

    Arguments:
        hidden_dim (int): number of dim in final output, does of projection of Value
        model_dim (int): number of dim in the attention
        key_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        key_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        query_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        query_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        value_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        value_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    query_linear = Dense(model_dim, init=query_init, init_bias=query_init_bias)
    key_linear = Dense(model_dim, init=key_init, init_bias=key_init_bias)
    value_linear = Dense(hidden_dim, init=value_init, init_bias=value_init_bias)

    def phi(x):  # kernel
        return C.elu(x) + 1

    @C.BlockFunction('LinearAttention', name=name)
    def model(query, key, value):
        q = phi(query_linear(query))
        k = phi(key_linear(key))
        v = value_linear(value)

        # key and value should have the same sequence length
        k_unpacked = C.sequence.unpack(k, padding_value=0, no_mask_output=True)
        # k_unpacked: [#] [*kv=, model_dim]
        v_unpacked = C.sequence.unpack(v, padding_value=0, no_mask_output=True)
        # v_unpacked: [#] [*kv=, hidden_dim]
        kv = C.times(C.swapaxes(k_unpacked), v_unpacked)
        # kv [#] [model_dim, hidden_dim]
        kv_broadcasted = C.sequence.broadcast_as(kv, q)  # this can be reused across queries
        # kv [#, *] [model_dim, hidden_dim]

        numerator = C.squeeze(C.times(C.expand_dims(q, axis=C.Axis.new_leading_axis()), kv_broadcasted))
        # numerator [#, *] [hidden_dim, ]
        denom = C.reduce_sum(q * C.sequence.broadcast_as(C.sequence.reduce_sum(k), q))
        # denom [#, *] [1]

        return numerator / denom

    return model


def LinearAttentionModel(hidden_dim: int, model_dim: int,
                         key_init=default_override_or(C.glorot_uniform()), key_init_bias=default_override_or(0),
                         query_init=default_override_or(C.glorot_uniform()), query_init_bias=default_override_or(0),
                         value_init=default_override_or(C.glorot_uniform()), value_init_bias=default_override_or(0),
                         name=''):
    """ Convenience wrapper in the style of cntk.layers.AttentionModel """
    attention = LinearAttention(hidden_dim=hidden_dim, model_dim=model_dim,
                                key_init=key_init, key_init_bias=key_init_bias,
                                query_init=query_init, query_init_bias=query_init_bias,
                                value_init=value_init, value_init_bias=value_init_bias, name=name)

    def model(encoder_hidden_state, decoder_hidden_state):
        return attention(decoder_hidden_state, encoder_hidden_state, encoder_hidden_state)

    return model


def ScaledDotProductAttention(obey_sequence_order: bool = None, max_seq_len: int = None, name=''):
    """
    Scaled dot-product attention implementation of "Attention is all you need", https://arxiv.org/abs/1706.03762

    An attention function can be described as mapping a query and a set of key-value pairs to an output,
    where the query, keys, values, and output are all vectors. The output is computed as a weighted sum
    of the values, where the weight assigned to each value is computed by a compatibility function of the
    query with the corresponding key.

    scaled_dot_product_attention(Q, K, V) = softmax(QV.T / sqrt(dk)) * V

    When query, key and value are all the same, it becomes self-attention.

    Note:
        Query and key must have the same dimension
        Key and value must have the same sequence length

    Example:
        a = C.sequence.input_variable(10)
        b = ScaledDotProductAttention()(a, a, a)

        assert b.shape == (10, )

    Arguments:
        obey_sequence_order: do not let attention peek into future values
        max_seq_len: max sequence length possible, used to ensure that sequence order is obeyed

    Returns:
        :class:`~cntk.ops.functions.Function`:
        A function that returns a weighted sum of value

    """

    def attention(query, key, value):
        return Cx.scaled_dot_product_attention(query, key, value, obey_sequence_order, max_seq_len, name)

    return attention


def MultiHeadAttention(num_heads, model_dim, obey_sequence_order: bool = None, max_seq_len: int = None,
                       key_init=default_override_or(C.glorot_uniform()), key_init_bias=default_override_or(0),
                       query_init=default_override_or(C.glorot_uniform()), query_init_bias=default_override_or(0),
                       value_init=default_override_or(C.glorot_uniform()), value_init_bias=default_override_or(0),
                       init=default_override_or(C.glorot_uniform()), init_bias=default_override_or(0),
                       name=''):
    """ Multi-head attention as described in "Attention is all you need", https://arxiv.org/abs/1706.03762

    Example:
        a = C.sequence.input_variable(10)
        b = MultiHeadAttention(2, 10)(a, a, a)

        assert b.shape == (10, )

    Arguments:
        num_heads (int): number of attention heads
        model_dim (int): number of hidden dim in final output of multi-head attention
        obey_sequence_order: do not let attention peek into future values
        max_seq_len: max sequence length possible, used to ensure that sequence order is obeyed
        key_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        key_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        query_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        query_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        value_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        value_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    assert model_dim % num_heads == 0, "Model dimension must be divisible by number of heads"

    head_dim = int(model_dim / num_heads)

    query_linear = Dense(model_dim, init=query_init, init_bias=query_init_bias)
    key_linear = Dense(model_dim, init=key_init, init_bias=key_init_bias)
    value_linear = Dense(model_dim, init=value_init, init_bias=value_init_bias)
    multihead_liner = Dense(model_dim, init=init, init_bias=init_bias)

    scaled_dot_product_attention = ScaledDotProductAttention(obey_sequence_order, max_seq_len)

    @C.BlockFunction('MultiHeadAttention', name)
    def inner(query, key, value):
        mixed_queries = query_linear(query)  # [#, *] {model_dim,]
        mixed_keys = key_linear(key)  # [#, *] {model_dim,]
        mixed_values = value_linear(value)  # [#, *] {model_dim,]

        # TODO: re-implement `ScaledDotProductAttention` when cntk has BatchMatMul so there's no need to slice here
        queries = [C.slice(mixed_queries, 0, i * head_dim, (i + 1) * head_dim) for i in range(num_heads)]
        keys = [C.slice(mixed_keys, 0, i * head_dim, (i + 1) * head_dim) for i in range(num_heads)]
        values = [C.slice(mixed_values, 0, i * head_dim, (i + 1) * head_dim) for i in range(num_heads)]

        # list of num_heads heads with shape (-3, head_dim) each
        attention_outputs = [scaled_dot_product_attention(q, k, v) for q, k, v in zip(queries, keys, values)]

        result = multihead_liner(C.splice(*attention_outputs))
        return result

    return _inject_name(inner, name)


def MultiHeadAttentionBlock(num_heads, model_dim, obey_sequence_order: bool = None, max_seq_len: int = None,
                            key_init=default_override_or(C.glorot_uniform()), key_init_bias=default_override_or(0),
                            query_init=default_override_or(C.glorot_uniform()), query_init_bias=default_override_or(0),
                            value_init=default_override_or(C.glorot_uniform()), value_init_bias=default_override_or(0),
                            init=default_override_or(C.glorot_uniform()), init_bias=default_override_or(0),
                            initial_scale=1, initial_bias=0, name=''):
    """ Multi head attention block as described in "Attention is all you need", https://arxiv.org/abs/1706.03762

    Multi-head attention block comes with a residual connection and a layer norm.

    Example:
        a = C.sequence.input_variable(10)
        b = MultiHeadAttentionBlock(2, 10)(a, a, a)

        assert b.shape == (10, )

    Arguments:
        num_heads (int): number of attention heads
        model_dim (int): number of hidden dim in final output of multi-head attention
        obey_sequence_order: do not let attention peek into future values
        max_seq_len: max sequence length possible, used to ensure that sequence order is obeyed
        key_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        key_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        query_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        query_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        value_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        value_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        initial_scale (float, default 1): initial value for the ``scale`` parameter aka gamma
        initial_bias (float, default 0): initial value for the ``bias`` parameter aka beta

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    attention_layer = MultiHeadAttention(num_heads, model_dim, obey_sequence_order, max_seq_len,
                                         key_init=key_init, key_init_bias=key_init_bias,
                                         query_init=query_init, query_init_bias=query_init_bias,
                                         value_init=value_init, value_init_bias=value_init_bias,
                                         init=init, init_bias=init_bias, name='MultiheadAttention')

    layernorm = LayerNormalization(initial_scale=initial_scale, initial_bias=initial_bias, name='LayerNorm')

    @C.Function
    def inner(query, key, value):
        attended = attention_layer(query, key, value)
        skip_connect_attended = attended + query
        normed_skip_connect_attended = layernorm(skip_connect_attended)
        return normed_skip_connect_attended

    return _inject_name(inner, name)


def TransformerEncoderBlock(num_heads: int, model_dim: int, intermediate_dim: int, dropout_rate: float = None,
                            obey_sequence_order: bool = None, max_seq_len: int = None,
                            key_init=default_override_or(C.glorot_uniform()), key_init_bias=default_override_or(0),
                            query_init=default_override_or(C.glorot_uniform()), query_init_bias=default_override_or(0),
                            value_init=default_override_or(C.glorot_uniform()), value_init_bias=default_override_or(0),
                            mha_init=default_override_or(C.glorot_uniform()), mha_init_bias=default_override_or(0),
                            mha_initial_scale=1, mha_initial_bias=0,
                            intermediate_init=default_override_or(C.glorot_uniform()), intermediate_init_bias=default_override_or(0),
                            init=default_override_or(C.glorot_uniform()), init_bias=default_override_or(0),
                            initial_scale=1, initial_bias=0, name=''):
    """ Encoder block of transformer as described in "Attention is all you need", https://arxiv.org/abs/1706.03762

    Consist of 1 multi head attention followed by a dense layer, residual connect and layer norm

    Arguments:
        num_heads (int): number of attention heads
        model_dim (int): number of hidden dim in final output of multi-head attention
        intermediate_dim (int): hidden/ intermediate dimension within position-wise feed-forward layer
        dropout_rate (float): probability of dropping out an element in the position-wise feed-forward
        obey_sequence_order: do not let attention peek into future values
        max_seq_len: max sequence length possible, used to ensure that sequence order is obeyed
        key_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        key_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        query_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        query_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        value_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        value_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        mha_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        mha_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
         mha_initial_scale (float, default 1): initial value for the ``scale`` parameter aka gamma
        mha_initial_bias (float, default 0): initial value for the ``bias`` parameter aka beta
        intermediate_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        intermediate_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        initial_scale (float, default 1): initial value for the ``scale`` parameter aka gamma
        initial_bias (float, default 0): initial value for the ``bias`` parameter aka beta

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    mha_block = MultiHeadAttentionBlock(num_heads, model_dim, obey_sequence_order, max_seq_len,
                                        key_init=key_init, key_init_bias=key_init_bias,
                                        query_init=query_init, query_init_bias=query_init_bias,
                                        value_init=value_init, value_init_bias=value_init_bias,
                                        init=mha_init, init_bias=mha_init_bias,
                                        initial_scale=mha_initial_scale, initial_bias=mha_initial_bias,
                                        name='SelfAttention')

    feed_foward = PositionwiseFeedForward(model_dim, intermediate_dim, dropout_rate=dropout_rate,
                                          intermediate_init=intermediate_init, intermediate_init_bias=intermediate_init_bias,
                                          init=init, init_bias=init_bias, name='PWFF')

    layernorm = LayerNormalization(initial_scale, initial_bias, name='LayerNorm')

    @C.Function
    def block(x):
        self_attended = mha_block(x, C.alias(x), C.alias(x))
        hidden = feed_foward(self_attended)
        output = layernorm(hidden + self_attended)  # residual connection
        return output

    return _inject_name(block, name)  # consider change to BlockFunction


def TransformerDecoderBlock(num_heads: int, model_dim: int, intermediate_dim: int, dropout_rate: float = None,
                            obey_sequence_order: bool = True, max_seq_len: int = None,
                            mha1_key_init=default_override_or(C.glorot_uniform()), mha1_key_init_bias=default_override_or(0),
                            mha1_query_init=default_override_or(C.glorot_uniform()), mha1_query_init_bias=default_override_or(0),
                            mha1_value_init=default_override_or(C.glorot_uniform()), mha1_value_init_bias=default_override_or(0),
                            mha1_init=default_override_or(C.glorot_uniform()), mha1_init_bias=default_override_or(0),
                            mha1_initial_scale=1, mha1_initial_bias=0,
                            mha2_key_init=default_override_or(C.glorot_uniform()), mha2_key_init_bias=default_override_or(0),
                            mha2_query_init=default_override_or(C.glorot_uniform()), mha2_query_init_bias=default_override_or(0),
                            mha2_value_init=default_override_or(C.glorot_uniform()), mha2_value_init_bias=default_override_or(0),
                            mha2_init=default_override_or(C.glorot_uniform()), mha2_init_bias=default_override_or(0),
                            mha2_initial_scale=1, mha2_initial_bias=0,
                            intermediate_init=default_override_or(C.glorot_uniform()),
                            intermediate_init_bias=default_override_or(0),
                            init=default_override_or(C.glorot_uniform()), init_bias=default_override_or(0),
                            initial_scale=1, initial_bias=0):
    """ Decoder block of transformer as described in "Attention is all you need", https://arxiv.org/abs/1706.03762

    Consist of 2 multi head attention followed by a dense layer, residual connect and layer norm

    Arguments:
        num_heads (int): number of attention heads
        model_dim (int): number of hidden dim in final output of multi-head attention
        intermediate_dim (int): hidden/ intermediate dimension within position-wise feed-forward layer
        dropout_rate (float): probability of dropping out an element in the position-wise feed-forward
        obey_sequence_order (bool, defaults True): do not let attention peek into future values
        max_seq_len (int): max sequence length possible, used to ensure that sequence order is obeyed
        mha1_key_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        mha1_key_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        mha1_query_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        mha1_query_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        mha1_value_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        mha1_value_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        mha1_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        mha1_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        mha1_initial_scale (float, default 1): initial value for the ``scale`` parameter aka gamma
        mha1_initial_bias (float, default 0): initial value for the ``bias`` parameter aka beta
        mha2_key_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        mha2_key_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        mha2_query_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        mha2_query_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        mha2_value_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        mha2_value_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        mha2_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        mha2_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        mha2_initial_scale (float, default 1): initial value for the ``scale`` parameter aka gamma
        mha2_initial_bias (float, default 0): initial value for the ``bias`` parameter aka beta
        intermediate_init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        intermediate_init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        init (scalar or NumPy array or :mod:`cntk.initializer`, defaults to :func:`~cntk.initializer.glorot_uniform` ): initial value of weights `W`
        init_bias (scalar or NumPy array or :mod:`cntk.initializer`, defaults to 0): initial value of weights `b`
        initial_scale (float, default 1): initial value for the ``scale`` parameter aka gamma
        initial_bias (float, default 0): initial value for the ``bias`` parameter aka beta

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    mha_block1 = MultiHeadAttentionBlock(num_heads=num_heads, model_dim=model_dim,
                                         obey_sequence_order=obey_sequence_order, max_seq_len=max_seq_len,
                                         key_init=mha1_key_init, key_init_bias=mha1_key_init_bias,
                                         query_init=mha1_query_init, query_init_bias=mha1_query_init_bias,
                                         value_init=mha1_value_init, value_init_bias=mha1_value_init_bias,
                                         init=mha1_init, init_bias=mha1_init_bias,
                                         initial_scale=mha1_initial_scale, initial_bias=mha1_initial_bias)
    
    mha_block2 = MultiHeadAttentionBlock(num_heads=num_heads, model_dim=model_dim,
                                         obey_sequence_order=False, max_seq_len=None,
                                         key_init=mha2_key_init, key_init_bias=mha2_key_init_bias,
                                         query_init=mha2_query_init, query_init_bias=mha2_query_init_bias,
                                         value_init=mha2_value_init, value_init_bias=mha2_value_init_bias,
                                         init=mha2_init, init_bias=mha2_init_bias,
                                         initial_scale=mha2_initial_scale, initial_bias=mha2_initial_bias)

    feed_foward = PositionwiseFeedForward(model_dim, intermediate_dim, dropout_rate=dropout_rate,
                                          intermediate_init=intermediate_init, intermediate_init_bias=intermediate_init_bias,
                                          init=init, init_bias=init_bias)

    layernorm = LayerNormalization(initial_scale, initial_bias)

    @C.Function
    def block(encoded, x):
        inner = mha_block1(x, x, x)
        inner = mha_block2(inner, encoded, encoded)
        output = layernorm(ResNetBlock(feed_foward)(inner))
        return output

    return block


def TransformerEncoder(n: int, num_heads: int, model_dim: int, intermediate_dim: int, dropout_rate: float = None):
    """ Transformer encoder as described in "Attention is all you need", https://arxiv.org/abs/1706.03762

    Example:
        a = C.sequence.input_variable(10)
        encoded = TransformerDecoder(3, 2, 10)(a)

        assert encoded.shape == (10, )

    Arguments:
        n (int): number of encoder blocks
        num_heads (int): number of attention heads
        model_dim (int): number of hidden dim in final output of multi-head attention
        intermediate_dim (int): hidden/ intermediate dimension within position-wise feed-forward layer
        dropout_rate (float): probability of dropping out an element in the position-wise feed-forward

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """

    blocks = [TransformerEncoderBlock(num_heads=num_heads, model_dim=model_dim, intermediate_dim=intermediate_dim,
                                      dropout_rate=dropout_rate, obey_sequence_order=False,
                                      max_seq_len=None) for __ in range(n)]

    @C.Function
    def inner(x):

        for block in blocks:
            x = block(x)

        return x

    return inner


def TransformerDecoder(n: int, num_heads: int, model_dim: int, intermediate_dim: int, dropout_rate: float = None,
                       max_seq_len: int = None):
    """ Transformer decoder as described in "Attention is all you need", https://arxiv.org/abs/1706.03762

    Example:
        a = C.sequence.input_variable(10)
        encoded = C.sequence.input_variable(10)

        decoded = TransformerDecoder(3, 2, 10)(encoded, a)

        assert decoded.shape == (10, )

    Arguments:
        n (int): number of decoder blocks
        num_heads (int): number of attention heads
        model_dim (int): number of hidden dim in final output of multi-head attention
        intermediate_dim (int): hidden/ intermediate dimension within position-wise feed-forward layer
        dropout_rate (float): probability of dropping out an element in the position-wise feed-forward
        max_seq_len: max sequence length possible, used to ensure that sequence order is obeyed

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """

    blocks = [TransformerDecoderBlock(num_heads=num_heads, model_dim=model_dim, intermediate_dim=intermediate_dim,
                                      dropout_rate=dropout_rate, obey_sequence_order=True, max_seq_len=max_seq_len)
              for __ in range(n)]

    @C.Function
    def decoder(encoded, x):

        for block in blocks:
            x = block(encoded, x)

        return x

    return decoder


def Transformer(num_encoder_blocks: int = 6, num_decoder_blocks=6, num_heads_encoder: int = 16,
                num_heads_decoder: int = 16, encoder_model_dim: int = 512, decoder_model_dim: int = 512,
                encoder_intermediate_dim: int = 2048, decoder_intermediate_dim: int = 2048,
                encoder_dropout_rate: float = 0.1, decoder_dropout_rate: float = 0.1,
                max_seq_len_decoder: int = 100):
    """ Transformer implementation as described in "Attention is all you need", https://arxiv.org/abs/1706.03762

    Example:
        a = C.sequence.input_variable(512)
        b = C.sequence.input_variable(512)

        transformer = Transformer()  # using default settings
        decoded = transformer(a, b)

        assert decoded.shape == (512, )

    Arguments:
        num_encoder_blocks: number of encoder blocks
        num_decoder_blocks: number of decoder blocks
        num_heads_encoder: number of encoder attention heads
        num_heads_decoder: number of decoder attention heads
        encoder_model_dim: encoder model output dimension (should be the same dimension as the transformer input)
        decoder_model_dim: decoder model output dimension (should be the same dimension as the transformer input)
        encoder_intermediate_dim (int): hidden/ intermediate dimension within position-wise feed-forward layer of encoder
        decoder_intermediate_dim (int): hidden/ intermediate dimension within position-wise feed-forward layer of decoder
        encoder_dropout_rate (float): probability of dropping out an element in the position-wise feed-forward of encoder
        decoder_dropout_rate (float): probability of dropping out an element in the position-wise feed-forward of decoder
        max_seq_len_decoder: max sequence length in decoding sequence. Used for preventing attention peeking into future values.

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    encoder = TransformerEncoder(n=num_encoder_blocks, num_heads=num_heads_encoder, model_dim=encoder_model_dim,
                                 intermediate_dim=encoder_intermediate_dim, dropout_rate=encoder_dropout_rate)

    decoder = TransformerDecoder(n=num_decoder_blocks, num_heads=num_heads_decoder, model_dim=decoder_model_dim,
                                 intermediate_dim=decoder_intermediate_dim, dropout_rate=decoder_dropout_rate,
                                 max_seq_len=max_seq_len_decoder)

    @C.Function
    def model(tensor_to_encode, decoder_input_tensor):
        encoded = encoder(tensor_to_encode)
        decoded = decoder(encoded, decoder_input_tensor)
        return decoded

    return model


def GaussianWindowAttention(nb_mixtures, activation=C.softplus, init=C.he_normal(), name=''):
    """
    Implementation of the attention model found in "Generating sequences with recurrent neural networks" by Alex Graves.

    Gaussian window attention uses a directional mixture of gaussian kernels as convolution/attention window.

    For more details, the paper can be found in https://arxiv.org/abs/1308.0850

    Note:
        There is a slight deviation from the original implementation where we use softplus as the activation
        function instead of exp. Exp activation causes some minor instability.
    
    Example:
        seq1 = C.Axis.new_unique_dynamic_axis('seq1')
        seq2 = C.Axis.new_unique_dynamic_axis('seq2')

        encoded = C.sequence.input_variable(30, sequence_axis=seq1)
        query = C.sequence.input_variable(28, sequence_axis=seq2)

        a = GaussianWindowAttention(10)(encoded, query)

        assert a.shape == (30, )

    Arguments:
        nb_mixtures (int): number of gaussian mixtures to use for attention model

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    dense = Dense(shape=3 * nb_mixtures, activation=activation, init=init, name="GravesAttention")

    def window_weight(a, b, k, u):
        """
        Calculate Phi is the window weight of character seq at position u of time t.
        Function tested to be correct on 2018-25-02 using numpy equivalent

        math:
            phi = summation of mixtures { a * exp ( -b * (k - u) ^ 2 ) }

        Args:
            a: importance of window within the mixture. Not normalised and doesn't sum to one.
            b: width of attention window
            k: location of window
            u: integer position of each item in sequence. Value from 1 to seq_length. (rank 2 tensor) [-3, 1]

        Returns:
            :class:`~cntk.ops.functions.Function`

        """
        # print(f"k shape: {k.shape}, u shape: {u.shape}")
        phi = a * C.exp(-1 * b * C.square(k - u))
        # print("internal phi shape:", phi.shape)
        phi = C.swapaxes(C.reduce_sum(phi, axis=0))  # Reduce sum the mixture axis
        # phi: [#, n] [*-c, 1]
        return phi

    @C.typemap
    def gaussian_windows_attention_coefficients(abk, nb_mixtures):
        """ Split into 3 equal tensor of dim nb_mixtures """
        a = C.slice(abk, 0, 0, nb_mixtures)
        b = C.slice(abk, 0, nb_mixtures, 2 * nb_mixtures)
        k = C.slice(abk, 0, 2 * nb_mixtures, 0)
        k = Recurrence(C.plus)(k)

        a = C.expand_dims(a, axis=-1)
        b = C.expand_dims(b, axis=-1)
        k = C.expand_dims(k, axis=-1)
        return a, b, k

    @C.BlockFunction('GaussianWindowAttention', name)
    def attention(encoded, network):
        abk = dense(network)
        a, b, k = gaussian_windows_attention_coefficients(abk, nb_mixtures)
        # print("abk shape:", a.shape, b.shape, k.shape)
        # a, b, k: [#, n] [nb_mixture, 1]
        # context: [#, c] [char_ohe]

        encoded_unpacked = C.sequence.unpack(encoded, padding_value=0, no_mask_output=True)
        # context_unpacked: [#] [*=c, char_ohe]
        u = Cx.sequence.position(encoded)  # position gives shape=(1, )
        # u: [#, c], [1]
        u_values, u_valid = C.sequence.unpack(u, padding_value=999_999).outputs
        # u_values: [#] [*=c, 1]
        # u_valid: [#] [*=c]
        u_values_broadcast = C.swapaxes(C.sequence.broadcast_as(u_values, k))
        # u_values_broadcast: [#, n] [1, *=c]
        u_valid_broadcast = C.sequence.broadcast_as(C.reshape(u_valid, (1,), 1), k)
        # u_valid_broadcast: [#, n] [*=c, 1] ~ shape verified correct at his point

        # print("u_values_broadcast shape:", u_values_broadcast.shape)
        # print("abk shape:", a.shape, b.shape, k.shape)
        phi = window_weight(a, b, k, u_values_broadcast)
        # phi: [#, n] [*=c, 1]
        zero = C.constant(0)
        phi = C.element_select(u_valid_broadcast, phi, zero, name="phi")
        # phi: [#, n] [*=c, 1]
        attended = C.reduce_sum(phi * C.sequence.broadcast_as(encoded_unpacked, phi), axis=0)
        # [#, n] [1, char_ohe]
        # print("attended_context shape:", attended_context.shape)
        output = C.squeeze(attended, name="GaussianWindowAttention")
        # [#, n] [char_ohe]
        return output

    return attention


def GaussianAttentionSeqImage(n: int, image_height: int, expected_image_width: int, name=''):
    """ Gaussian attention applied to an encoded sequence image (i.e. sequence axis is image width)

    This implementation is from the deepmind paper, DRAW: A Recurrent Neural Network for Image Generation by Gregor et al
    More details can be found in the following https://arxiv.org/abs/1502.04623

    Example:
        n = 5
        num_channels = 3
        image_height = 64
        expected_image_width = 1000
        image_seq = C.sequence.input_variable((num_channels, image_height))  # rgb image with variable width and fixed height
        decoder_hidden_state = ...  # from decoder somewhere in the network
        attended_image = Cx.layers.GaussianAttentionSeqImage(n, image_height, expected_image_width)(image_seq, decoder_hidden_state)

        assert attended_image.shape == (num_channels, n, n)

    Arguments:
        n (int): number of gaussian attention filter per grid dimension,
          where total of number of attention filter = n * n grid
        image_height (int): the static image height dimension of the sequence
        expected_image_width (int): Expected number of cols (width) in the image

    """
    dense = Dense(shape=(5, ))
    A = expected_image_width
    B = image_height

    def attention_parameters(network_outputs):
        g_x = 0.5 * (A + 1) * (network_outputs[0] + 1)  # grid centre - x (cols)
        g_y = 0.5 * (B + 1) * (network_outputs[1] + 1)  # grid centre - y (rows)
        sigma2 = C.exp(network_outputs[2])  # isotropic variance
        delta = (max(A, B) - 1) / (n - 1) * C.exp(network_outputs[3])  # stride
        gamma = C.exp(network_outputs[4])  # intensity
        return g_x, g_y, sigma2, delta, gamma

    @C.BlockFunction('GaussianAttentionSeqImage', name)
    def model(seq_image, decoded):
        params = dense(decoded)
        g_x, g_y, sigma2, delta, gamma = attention_parameters(params)

        i = C.Constant(np.arange(n) + 1,)  # col of patch
        j = C.Constant(np.arange(n) + 1,)  # row of patch
        mu_x = g_x + (i - n / 2 - 0.5) * delta
        mu_y = g_y + (j - n / 2 - 0.5) * delta
        mu_x = C.expand_dims(mu_x, axis=-1)
        mu_y = C.expand_dims(mu_y, axis=-1)
        # mu_x: [#, *] [n, 1]
        # mu_y: [#, *] [n, 1]

        image = C.sequence.unpack(seq_image, padding_value=0, no_mask_output=True)
        # image: [#] [*image_width, filters, image_height]

        width_pos = Cx.sequence.position(seq_image)
        # width_pos: [#, *] [1]

        width_pos_unpacked = C.sequence.unpack(width_pos, padding_value=999_999, no_mask_output=True)
        # width_pos: [#] [*image_width, 1]

        a = C.sequence.broadcast_as(C.swapaxes(width_pos_unpacked), mu_x)
        # a: [#, *] [1, *image_width]
        # x pos index of image (width)

        b = C.Constant(np.arange(image_height).reshape((1, -1)))
        # b: [] [1, image_height]
        # y pos index of image (height)

        # calculate the which portion of the image that is attended by the gaussian filter
        f_xi = C.exp(-0.5 * C.square(a - mu_x) / sigma2)
        f_yj = C.exp(-0.5 * C.square(b - mu_y) / sigma2)
        # f_xi: [#, *] [n, *image_width]
        # f_yj: [#, *] [n, image_height]

        z_x = C.reduce_sum(f_xi, axis=1)
        z_y = C.reduce_sum(f_yj, axis=1)
        # z_x: [#, *] [n]
        # z_y: [#, *] [n]

        f_xi = f_xi / z_x
        f_yj = f_yj / z_y
        # f_xi: [#, *] [n, *image_width]
        # f_yj: [#, *] [n, image_height]

        # combine filters from x and y
        image_broadcasted = C.sequence.broadcast_as(image, f_yj)
        attended = gamma * C.times(f_xi, C.times_transpose(image_broadcasted, f_yj), output_rank=2)
        # attended: [#, *] [n, filters, n]
        attended = C.swapaxes(attended)
        # attended: [#, *] [filters, n (x) , n (y)]
        return attended

    return model


def PreTrainedBertEncoder(tf_bert_model_filepath: str, num_heads: int, dropout_rate: float = None):
    """ Use pre-trained tensorflow bert model

    Currently it is tested to work with:
        - `BERT-Base, Uncased`, uncased_L-12_H-768_A-12

    Models can be downloaded at https://github.com/google-research/bert

    Arguments:
        tf_bert_model_filepath (str): file path to the tensorflow model
        num_heads (int): number of attention heads in self attention
        dropout_rate (float): probability of dropping out an element in encoder

    Returns:
        :class:`~cntk.ops.functions.Function`:
        TF to CNTK Pre-trained Bert Encoder (Transformer Encoder)
    """
    try:
        import tensorflow as tf
    except ImportError:
        raise ImportError("Loading a TensorFlow models in CNTK, requires TensorFlow to be installed. Please see "
                          "https://www.tensorflow.org/install/ for installation instructions.")

    def bert_encoder_layer_number(layer_name: str, prefix):
        """ extracts 'xx' in '{prefix}{layer_xx/}{rest of the layer name}'

        There must not be any '/' on the left of layer_xx after prefix has been ignored.

        Arguments:
            layer_name (str): name of the layer
            prefix (str): prefix to be ignored

        Returns:
            int

        """
        # TODO: use regex probably more straightforward
        end_index = layer_name.index('/', len(prefix))
        number = int(layer_name[len(prefix):end_index].replace('layer_', ''))
        return number

    bert_encoder_prefix = 'bert/encoder/'

    variables_meta = tf.train.list_variables(tf_bert_model_filepath)
    encoder_variable_meta = [meta for meta in variables_meta if bert_encoder_prefix in meta[0]]

    layer_numbers = [bert_encoder_layer_number(meta[0], bert_encoder_prefix) for meta in encoder_variable_meta]
    nb_layers = max(layer_numbers) + 1  # +1 because layer numbering assumed to start from zero
    assert min(layer_numbers) == 0, f"Layer numbering assumed to start from zero but loaded model start from {min(layer_numbers)}"

    intermediate_dim = [meta[1][0] for meta in encoder_variable_meta if 'intermediate/dense/bias' in meta[0]]
    assert all(dim == intermediate_dim[0] for dim in intermediate_dim)
    intermediate_dim = intermediate_dim[0]

    model_dim = [meta[1][0] for meta in encoder_variable_meta if 'attention/output/dense/bias' in meta[0]]
    assert all(dim == model_dim[0] for dim in model_dim)
    model_dim = model_dim[0]

    encoder_layers = []
    mha_output_layernorm_bias_tag = 'attention/output/LayerNorm/beta'
    mha_output_layernorm_scale_tag = 'attention/output/LayerNorm/gamma'

    mha_output_dense_bias_tag = 'attention/output/dense/bias'
    mha_output_dense_kernel_tag = 'attention/output/dense/kernel'
    
    mha_key_bias_tag = 'attention/self/key/bias'
    mha_key_kernel_tag = 'attention/self/key/kernel'
    mha_query_bias_tag = 'attention/self/query/bias'
    mha_query_kernel_tag = 'attention/self/query/kernel'
    mha_value_bias_tag = 'attention/self/value/bias'
    mha_value_kernel_tag = 'attention/self/value/kernel'

    mha_dense_bias_tag = 'intermediate/dense/bias'
    mha_dense_kernel_tag = 'intermediate/dense/kernel'

    output_dense_bias_tag = 'output/dense/bias'
    output_dense_kernel_tag = 'output/dense/kernel'
    output_layernorm_scale_tag = 'output/LayerNorm/gamma'
    output_layernorm_bias_tag = 'output/LayerNorm/beta'

    config = {'num_heads': num_heads,
              'model_dim': model_dim,
              'intermediate_dim': intermediate_dim,
              'dropout_rate': dropout_rate,
              'obey_sequence_order': False,
              'max_seq_len': None,
              'key_init': mha_key_kernel_tag,
              'key_init_bias': mha_key_bias_tag,
              'query_init': mha_query_kernel_tag,
              'query_init_bias': mha_query_bias_tag,
              'value_init': mha_value_kernel_tag,
              'value_init_bias': mha_value_bias_tag,
              'mha_init': mha_output_dense_kernel_tag,
              'mha_init_bias': mha_output_dense_bias_tag,
              'mha_initial_scale': mha_output_layernorm_scale_tag,
              'mha_initial_bias': mha_output_layernorm_bias_tag,
              'intermediate_init': mha_dense_kernel_tag,
              'intermediate_init_bias': mha_dense_bias_tag,
              'init': output_dense_kernel_tag,
              'init_bias': output_dense_bias_tag,
              'initial_scale': output_layernorm_scale_tag,
              'initial_bias': output_layernorm_bias_tag,
              'name': None}
    
    for layer_num in range(nb_layers):
        prefix = f'bert/encoder/layer_{layer_num}/'
        initialised_config = {k: tf.train.load_variable(tf_bert_model_filepath, prefix + v) if isinstance(v, str) else v
                              for k, v in config.items()}

        initialised_config['name'] = f'encoder_layer_{layer_num}'
        encoder_layers.append(TransformerEncoderBlock(**initialised_config))

    @C.Function
    def model(x):

        for encoder_layer in encoder_layers:
            x = encoder_layer(x)

        return x

    return _inject_name(model, 'bert')


def PreTrainedBertModel(tf_bert_model_filepath: str, num_heads: int, dropout_rate: float = None):
    """ Initialise a pre-trained CNTK bert model converted from tensorflow

    Currently it is tested to work with:
        - `BERT-Base, Uncased`, uncased_L-12_H-768_A-12

    Models can be downloaded at https://github.com/google-research/bert

    Arguments:
        tf_bert_model_filepath (str): file path to the tensorflow model
        num_heads (int): number of attention heads in self attention
        dropout_rate (float): probability of dropping out an element in embedding and encoder

    Returns:
        :class:`~cntk.ops.functions.Function`:
        TF to CNTK Pre-trained Bert Model
    """
    try:
        import tensorflow as tf
    except ImportError:
        raise ImportError("Loading a TensorFlow models in CNTK, requires TensorFlow to be installed. Please see "
                          "https://www.tensorflow.org/install/ for installation instructions.")

    bert_embeddings = PreTrainedBertEmbeddings(tf_bert_model_filepath, dropout_rate)
    bert_encoder = PreTrainedBertEncoder(tf_bert_model_filepath, num_heads, 0.1)
    bert_pooler = PretrainedBertPooler(tf_bert_model_filepath)

    @C.Function
    def model(text_tensor, token_type_tensor):
        embedded = bert_embeddings(text_tensor, token_type_tensor)
        encoded = bert_encoder(embedded)
        pooled = bert_pooler(encoded)  # pooled is no longer a cntk sequence
        return pooled

    return model
